from interface.network.port_forwarding import PortForwardingInterface
from interface.network.resources import Port, FloatingIP, PortForwarding

from implementation.utils.connection import OpenStackConnection
from implementation.utils.decorators import OpenStackError

import random

# for limiting the ones that can be allocated over floating ips
RESERVED_PORT_NUMBERS = [22, 80, 9696]


class OSPortForwardingManager(PortForwardingInterface, OpenStackConnection):

    # TODO: gestionar els rangs de ports lliures/ocupats i floating ips? aquí o al django o on?
    # el delete el podem fer a partir de la ip i no de la instance_id


    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.floating_ips = [
            FloatingIP(item) for item in self.openstack.list_floating_ips()]

        for floating_ip in self.floating_ips:
            items = self.openstack.network.floating_ip_port_forwardings(floating_ip.id)
            floating_ip.port_forwardings = [
                PortForwarding(item, floating_ip.address) for item in items]

        #self.port_forwardings = [
        #    PortForwarding(item) for item in self.openstack.network.port_forwardings(self.floating_ips[0].id)]


    @OpenStackError
    def create_port_forwarding(self, instance_id, internal_port_number, protocol):
        """DUBTE hem de gestionar condicions com ara que no estigui creat ja
        un port forwarding igual, que la floating_ip no estigui activa, que no
        estigui associat a un router, etc.?
        Les floating ips estan separades de la ip on hi ha les apis dels serveis
        d'openstack? Perquè en devstack no, i no es pot habilitar un port extern
        que ja es faci servir (s'ha de restringir)?
        """

        project_id = self.openstack.auth['project_id']

        # fetch internal Neutron port id of the instance by instance_id
        internal_port_object = self.__get_internal_port_object_from_instance_id(instance_id)
        internal_port_id = internal_port_object.id
        internal_ip_address = internal_port_object.ip_address

        # TODO: recorrer cada floating ip i cridar una funció que ens torni un port lliure o None, sinó crear nova ip
        floating_ip_id = self.floating_ips[0].id
        floating_ip_address = self.floating_ips[0].address

        external_port = self.__obtain_free_port_by_floating_ip_id(floating_ip_id)
        print(f"PORT {external_port}")

        return PortForwarding(self.openstack.network.create_port_forwarding(
            floatingip_id=floating_ip_id,
            internal_port_id=internal_port_id,
            internal_ip_address=internal_ip_address,
            internal_port=int(internal_port_number),
            external_port=external_port,
            protocol=protocol,
            description=''), floating_ip_address)

        # more handcrafted alternatives

        #res = self.openstack.network.post(
        #    f"/floatingips/{floating_ip_id}/port_forwardings", json=payload)

        #r = requests.post("http://vm.niif.cloud.bme.hu:18686/v2.0/floatingips" \
        #    f"/{floating_ip_id}/port_forwardings", data = payload)
        #r.json()


    @OpenStackError
    def delete_port_forwarding(self, instance_id, internal_port_number):
        """ Removes port forwarding rule
        """

        # fetch internal Neutron port of the instance by instance_id
        internal_port_object = self.__get_internal_port_object_from_instance_id(instance_id)
        internal_port_id = internal_port_object.id
        internal_ip_address = internal_port_object.ip_address

        # find floating ip where port forwarding rule is defined
        floating_ip_id = None
        for floating_ip in self.openstack.list_floating_ips():
            for port_forwarding in floating_ip.port_forwardings:
                if ((port_forwarding.internal_ip_address == internal_ip_address)
                    and (port_forwarding.internal_port == internal_port_number)):
                    # found
                    floating_ip_id = floating_ip.id
                    break
                    break

        # loop again to fetch port forwarding id
        # because is not contained in objects from the above results
        port_forwarding_id = None
        for port_forwarding in self.openstack.network.floating_ip_port_forwardings(floating_ip_id):
            if ((port_forwarding.internal_ip_address == internal_ip_address)
                and (port_forwarding.internal_port == internal_port_number)):
                port_forwarding_id = port_forwarding.id

        self.openstack.network.delete_port_forwarding(
            port_forwarding_id, floating_ip_id)


    @OpenStackError
    def list_instance_port_forwardings(self, instance_id):
        """ Lists all port forwarding rules defined from a vm instance
        """

        # fetch internal Neutron port id of the instance by instance_id
        internal_port_object = self.__get_internal_port_object_from_instance_id(instance_id)
        internal_port_id = internal_port_object.id
        internal_ip_address = internal_port_object.ip_address

        # fetch port forwardings
        instance_port_forwardings = []
        for floating_ip in self.floating_ips:
            for item in self.openstack.network.port_forwardings(floating_ip.id):
                if item.internal_ip_address == internal_ip_address:
                    instance_port_forwardings.append(PortForwarding(item, floating_ip.address))

        return instance_port_forwardings


    def __get_internal_port_object_from_instance_id(self, instance_id):
        """ Fetches Neutron internal port object from instance_id
        """

        # query ports filtered by keyboard arguments (kwargs), returns a list
        results = [Port(item) for item in self.openstack.network.ports(
            device_id=instance_id)]

        # pick the result, we assume the list lenght is 1
        internal_port_object = results.pop()

        return internal_port_object


    def __obtain_free_port_by_floating_ip_id(self, floating_ip_id):
        """ Pot una vm tenir port forwardings en diferents floating ips?
        """

        # initialize list with the full range of ports
        port_numbers = list(range(1, 2**16))

        # remove reserved ones
        for reserved in RESERVED_PORT_NUMBERS:
            if reserved in port_numbers:
                port_numbers.remove(reserved)

        # query all current port forwardings from the floating ip
        already_allocated_port_numbers = []
        for item in self.openstack.network.port_forwardings(floating_ip_id):
            already_allocated_port_numbers.append(PortForwarding(item).external_port)

        # remove already allocated ports from range list
        for allocated in already_allocated_port_numbers:
            if allocated in port_numbers:
                port_numbers.remove(allocated)

        free_port_number = None
        if len(port_numbers):
            free_port_number = random.choice(port_numbers)

        return free_port_number
