from openstack.exceptions import SDKException
from novaclient import client
from interface_openstack.implementation.utils.connection import OpenStackConnection
from interface_openstack.implementation.utils.decorators import OpenStackError
import secrets

from interface_openstack.interface.vm.instance import InstanceInterface
from interface_openstack.interface.vm.resources import Instance, Flavor
from interface_openstack.interface.image.image import Image


class OSVirtualMachineManager(InstanceInterface, OpenStackConnection):

    @OpenStackError
    def _get_openstack_server(self, name_or_id):
        if not name_or_id:
            raise ValueError("Name or id doesn't given")
        server_instance = self.openstack.compute.get_server(name_or_id)
        if not server_instance:
            raise ValueError("Could not get the vm")
        return server_instance

    @OpenStackError
    def create_base_vm(self, name, flavor, networks, block_dev_map):
        devices = []
        b_device = block_dev_map.__dict__
        devices.append(b_device)
        flavor = self.get_flavor(flavor)
        new_server = self.openstack.compute.create_server(
                 name=name,
                 flavorRef=flavor.id,
                 networks=networks,
                 block_device_mapping=devices
                 )
        return self.convert_server_to_instance(new_server)

    @OpenStackError
    def create_vm_from_template(self, name, image, flavor, networks):
        return self.create_multiple_vm_from_template(name, image, flavor,
                                                     networks, 1)[0]

    @OpenStackError
    def create_multiple_vm_from_template(self, name, image, flavor, networks,
                                         number, **args):
        compute = self.openstack.compute
        flav = compute.find_flavor(flavor)

        image = compute.find_image(image)
        if not image:
            raise ValueError("The template not found")

        server_ids = []

        # TODO: Maybe we need parallel execution of multiple creation
        for i in range(0, number):
            new_server = compute.create_server(name=name,
                                               flavorRef=flav.id,
                                               imageRef=image.id,
                                               networks=networks,
                                               )
            server_ids.append(new_server.id)

        return server_ids

    @OpenStackError
    def create_flavor(self, name, ram, vcpus, initial_disk):
        flavor = self.openstack.compute.create_flavor(name=name,
                                                      ram=ram,
                                                      vcpus=vcpus,
                                                      disk=initial_disk)
        return Flavor(flavor.name, flavor.id, flavor.ram,
                      flavor.vcpus, flavor.disk)

    @OpenStackError
    def get_flavor(self, flavor_id):
        flavor = self.openstack.compute.find_flavor(flavor_id)
        return Flavor(flavor.name, flavor.id, flavor.ram,
                      flavor.vcpus, flavor.disk)

    @OpenStackError
    def delete_flavor(self, flavor_id):
        flavor = self.openstack.compute.find_flavor(flavor_id)
        self.openstack.compute.delete_flavor(flavor)

    @OpenStackError
    def list_flavors(self):
        flavors = []
        for flavor in self.openstack.compute.flavors():
            flavors.append(Flavor(flavor.name, flavor.id, flavor.ram,
                                  flavor.vcpus, flavor.disk))
        return flavors

    @OpenStackError
    def get_vm(self, name_or_id=None):
        instance = self._get_openstack_server(name_or_id)
        return self.convert_server_to_instance(instance)

    @OpenStackError
    def start_vm(self, name_or_id=None):
        instance = self._get_openstack_server(name_or_id)
        self.openstack.compute.start_server(instance)

    @OpenStackError
    def stop_vm(self, name_or_id=None):
        instance = self._get_openstack_server(name_or_id)
        self.openstack.compute.stop_server(instance)

    @OpenStackError
    def suspend_vm(self, name_or_id=None):
        instance = self._get_openstack_server(name_or_id)
        self.openstack.compute.suspend_server(instance)

    @OpenStackError
    def wake_up_vm(self, name_or_id=None):
        instance = self._get_openstack_server(name_or_id)
        self.openstack.compute.resume_server(instance)

    @OpenStackError
    def reboot_vm(self, name_or_id):
        instance = self._get_openstack_server(name_or_id)
        self.openstack.compute.reboot_server(instance, reboot_type='SOFT')

    @OpenStackError
    def reset_vm(self, name_or_id):
        instance = self._get_openstack_server(name_or_id)
        self.openstack.compute.reboot_server(instance, reboot_type='HARD')

    @OpenStackError
    def destroy_vm(self, name_or_id):
        instance = self._get_openstack_server(name_or_id)
        self.openstack.compute.delete_server(instance)

    @OpenStackError
    def get_status(self, name_or_id):
        instance = self._get_openstack_server(name_or_id)
        return instance.status

    @OpenStackError
    def list_all_vm(self):
        servers = []
        for server in self.openstack.compute.servers():
            servers.append(self.convert_server_to_instance(server))
        return servers

    @OpenStackError
    def resize_vm(self, name_or_id, resource):
        instance = self._get_openstack_server(name_or_id)
        flavor = self.openstack.compute.find_flavor(resource['name'])

        self.openstack.compute.resize_server(instance, flavor)

    @OpenStackError
    def create_image(self, instance_id, metadata=None):
        image_name = secrets.token_hex(30)
        self.openstack.compute.create_server_image(instance_id,
                                                   image_name,
                                                   metadata)
        image = self.openstack.compute.find_image(image_name)
        return  self.os_image_to_rc_image(image)

    def get_vnc_console(self, server_id):
        with client.Client("2", session=self.client_session) as nova:
            if server_id:
                instance = nova.servers.get(server_id)
                return instance.get_vnc_console("novnc")

    @OpenStackError
    def attach_volume(self, server_id, volume_id, device=None):
        self.openstack.compute.create_volume_attachment(volumeId=volume_id,
                                                        server=server_id,
                                                        )

    @OpenStackError
    def detach_volume(self, server_id, volume_id, device=None):
        self.openstack.compute.delete_volume_attachment(
            volume_attachment=volume_id,
            server=server_id)

    def convert_server_to_instance(self, server):
        if not server.image:
            image_id = None
        else:
            image_id = server.image["id"]
        return Instance(id=server.id,
                        flavor=server.flavor["id"],
                        name=server.name,
                        image_id=image_id,
                        disks=server.attached_volumes,
                        status=server.status,
                        launched_at=server.launched_at,
                        terminated_at=server.terminated_at,
                        addresses=server.addresses)

    def os_image_to_rc_image(self, os_image):
        return Image(
            os_image.id,
            os_image.name
        )
