import json


class Port:
    """ From the OpenStack Port source data:
    - Releveant attributes are stored
    - Some methods regarding class representation
    """

    def __init__(self, data):
        self.id = data['id'] # str
        self.ip_address = None # see below
        self.ipv6_address = None # see below
        self.status = data['status'] # str e.g. 'ACTIVE'
        self.device_id = data['device_id'] # str
        self.created_at = data['created_at'] # str
        self.updated_at = data['updated_at'] # str

        if len(data['fixed_ips']):
            for ip in data['fixed_ips']:
                if ':' in ip['ip_address']: # ipv6
                    self.ipv6_address = data['fixed_ips'][1]['ip_address'] # str
                else:
                    self.ip_address = data['fixed_ips'][0]['ip_address'] # str

    def __repr__(self):
        return f"Port({self.ip_address})"

    def __str__(self):
        return f"Port({self.ip_address})"

    def JSON(self):
        return json.dumps(self.__dict__)


class FloatingIP:
    """ From the OpenStack Floating IP source data:
    - Releveant attributes are stored
    - Some methods regarding class representation
    """

    def __init__(self, data):
        self.id = data['id'] # str
        self.address = data['floating_ip_address'] # str
        self.status = data['status'] # str e.g. 'ACTIVE'
        self.floating_network_id = data['floating_network_id'] # str
        self.router_id = data['router_id'] # str
        self.port_forwardings = [] # objects come w/o id here, are fetched after
        # [PortForwarding(item) for item in data['port_forwardings']]
        self.created_at = data['created_at'] # str
        self.updated_at = data['updated_at'] # str

    def __repr__(self):
        return f"FloatingIP({self.address})"

    def __str__(self):
        return f"FloatingIP({self.address})"

    def JSON(self):
        return json.dumps(self.__dict__)


class PortForwarding:
    """ From the OpenStack Port Forwarding source data:
    - Releveant attributes are stored
    - Some methods regarding class representation
    """

    def __init__(self, data, floating_ip_address=''):
        self.id = data['id'] # str
        self.protocol = data['protocol'] # str
        self.internal_ip_address = data['internal_ip_address'] # str
        self.external_ip_address = floating_ip_address # str
        self.internal_port = data['internal_port'] # number
        self.external_port = data['external_port'] # number

    def __repr__(self):
        return(
            f"PortForwarding({self.internal_ip_address}:{self.internal_port}"
            f" -> {self.external_ip_address}:{self.external_port})")

    def __str__(self):
        return(
            f"PortForwarding({self.internal_ip_address}:{self.internal_port}"
            f" -> {self.external_ip_address}:{self.external_port})")

    def JSON(self):
        return json.dumps(self.__dict__)


class Interface:
    """ From the OpenStack Network source data, which is the equivalent concept
    for the Circle "interface". In OpenStack, to achieve a "vlan", is required
    a specific Router where the Network is attached to, and two Subnets inside
    of the Network, one for IPv4 and one for IPv6. Also:
    - Some methods regarding class representation
    """

    def __init__(self, data):
        self.id = data['id'] # str
        self.name = data['name'] # str
        self.port = None # Port see below
        self.is_external = data['is_router_external'] # bool
        self.created_at = data['created_at'] # str
        self.updated_at = data['updated_at'] # str

    def set_port(self, data):
        """ This attribute is set when a device is added to the interface, not
        when the interface is created.

        Specifies the "connection" from the device to the network
        """
        self.port = Port(data)

    def __repr__(self):
        return(f"Interface({self.name})")

    def __str__(self):
        return(f"Interface({self.name})")

    def JSON(self):
        return json.dumps(self.__dict__)
